@@ -1,7 +1,7 @@
 /*
  * drivers/video/tegra/dc/dev.c
  *
- * Copyright (C) 2011, NVIDIA Corporation
+ * Copyright (C) 2011-2012, NVIDIA Corporation
  *
  * Author: Robert Morell <rmorell@nvidia.com>
  * Some code based on fbdev extensions written by:
@@ -29,7 +29,6 @@
 #include <mach/dc.h>
 #include <mach/nvmap.h>
 #include <mach/tegra_dc_ext.h>
-#include <mach/fb.h>
 
 /* XXX ew */
 #include "../dc_priv.h"
@@ -45,9 +44,10 @@ static int head_count;
 
 struct tegra_dc_ext_flip_win {
 	struct tegra_dc_ext_flip_windowattr	attr;
-	struct nvmap_handle_ref			*handle;
-	/* ugh. is this really necessary */
+	struct nvmap_handle_ref			*handle[TEGRA_DC_NUM_PLANES];
 	dma_addr_t				phys_addr;
+	dma_addr_t				phys_addr_u;
+	dma_addr_t				phys_addr_v;
 	u32					syncpt_max;
 };
 
@@ -140,17 +140,15 @@ static void set_enable(struct tegra_dc_ext *ext, bool en)
 	 * Take all locks to make sure any flip requests or cursor moves are
 	 * out of their critical sections
 	 */
-	for (i = 0; i < ext->dc->n_windows; i++) {
+	for (i = 0; i < ext->dc->n_windows; i++)
 		mutex_lock(&ext->win[i].lock);
-	}
 	mutex_lock(&ext->cursor.lock);
 
 	ext->enabled = en;
 
 	mutex_unlock(&ext->cursor.lock);
-	for (i = ext->dc->n_windows - 1; i >= 0 ; i--) {
+	for (i = ext->dc->n_windows - 1; i >= 0 ; i--)
 		mutex_unlock(&ext->win[i].lock);
-	}
 }
 
 void tegra_dc_ext_enable(struct tegra_dc_ext *ext)
@@ -180,9 +178,9 @@ static int tegra_dc_ext_set_windowattr(struct tegra_dc_ext *ext,
 {
 	struct tegra_dc_ext_win *ext_win = &ext->win[win->idx];
 
-	if (flip_win->handle == NULL) {
+	if (flip_win->handle[TEGRA_DC_Y] == NULL) {
 		win->flags = 0;
-		ext_win->cur_handle = NULL;
+		memset(ext_win->cur_handle, 0, sizeof(ext_win->cur_handle));
 		return 0;
 	}
 
@@ -191,15 +189,12 @@ static int tegra_dc_ext_set_windowattr(struct tegra_dc_ext *ext,
 		win->flags |= TEGRA_WIN_FLAG_BLEND_PREMULT;
 	else if (flip_win->attr.blend == TEGRA_DC_EXT_BLEND_COVERAGE)
 		win->flags |= TEGRA_WIN_FLAG_BLEND_COVERAGE;
-
-	/*
-	 * If we start with a swap_countdown of zero, that means we want to
-	 * swap as soon as possible.  Mark this window as such, because it
-	 * affects what interrupt we use to trigger the swap completion.
-	 */
-	if (win->swap_countdown == 0)
-		win->flags |= TEGRA_WIN_FLAG_SWAP_ASAP;
-
+	if (flip_win->attr.flags & TEGRA_DC_EXT_FLIP_FLAG_TILED)
+		win->flags |= TEGRA_WIN_FLAG_TILED;
+	if (flip_win->attr.flags & TEGRA_DC_EXT_FLIP_FLAG_INVERT_H)
+		win->flags |= TEGRA_WIN_FLAG_INVERT_H;
+	if (flip_win->attr.flags & TEGRA_DC_EXT_FLIP_FLAG_INVERT_V)
+		win->flags |= TEGRA_WIN_FLAG_INVERT_V;
 	win->fmt = flip_win->attr.pixformat;
 	win->x.full = flip_win->attr.x;
 	win->y.full = flip_win->attr.y;
@@ -211,54 +206,52 @@ static int tegra_dc_ext_set_windowattr(struct tegra_dc_ext *ext,
 	win->out_w = flip_win->attr.out_w;
 	win->out_h = flip_win->attr.out_h;
 	win->z = flip_win->attr.z;
-	ext_win->cur_handle = flip_win->handle;
+	memcpy(ext_win->cur_handle, flip_win->handle,
+	       sizeof(ext_win->cur_handle));
 
 	/* XXX verify that this won't read outside of the surface */
 	win->phys_addr = flip_win->phys_addr + flip_win->attr.offset;
-	win->offset_u = flip_win->attr.offset_u + flip_win->attr.offset;
-	win->offset_v = flip_win->attr.offset_v + flip_win->attr.offset;
+
+	win->phys_addr_u = flip_win->handle[TEGRA_DC_U] ?
+		flip_win->phys_addr_u : flip_win->phys_addr;
+	win->phys_addr_u += flip_win->attr.offset_u;
+
+	win->phys_addr_v = flip_win->handle[TEGRA_DC_V] ?
+		flip_win->phys_addr_v : flip_win->phys_addr;
+	win->phys_addr_v += flip_win->attr.offset_v;
+
 	win->stride = flip_win->attr.stride;
 	win->stride_uv = flip_win->attr.stride_uv;
 
 	if ((s32)flip_win->attr.pre_syncpt_id >= 0) {
-		nvhost_syncpt_wait_timeout(&ext->dc->ndev->host->syncpt,
-					   flip_win->attr.pre_syncpt_id,
-					   flip_win->attr.pre_syncpt_val,
-					   msecs_to_jiffies(500));
+		nvhost_syncpt_wait_timeout(
+				&nvhost_get_host(ext->dc->ndev)->syncpt,
+				flip_win->attr.pre_syncpt_id,
+				flip_win->attr.pre_syncpt_val,
+				msecs_to_jiffies(500), NULL);
 	}
 
 
 	return 0;
 }
 
-static void process_window_change(struct tegra_dc_ext *ext, int delta)
-{
-	mutex_lock(&ext->enable_change_lock);
-
-	if (!ext->nr_win_ena || ext->nr_win_ena + delta == 0)
-		tegra_fb_transition(ext->dc->fb, delta < 0);
-
-	ext->nr_win_ena += delta;
-	BUG_ON(ext->nr_win_ena < 0);
-
-	mutex_unlock(&ext->enable_change_lock);
-}
-
 static void tegra_dc_ext_flip_worker(struct work_struct *work)
 {
 	struct tegra_dc_ext_flip_data *data =
 		container_of(work, struct tegra_dc_ext_flip_data, work);
 	struct tegra_dc_ext *ext = data->ext;
 	struct tegra_dc_win *wins[DC_N_WINDOWS];
-	struct nvmap_handle_ref *unpin_handles[DC_N_WINDOWS];
-	int i, nr_unpin = 0, nr_win = 0, nr_disable = 0;
+	struct nvmap_handle_ref *unpin_handles[DC_N_WINDOWS *
+					       TEGRA_DC_NUM_PLANES];
+	struct nvmap_handle_ref *old_handle;
+	int i, nr_unpin = 0, nr_win = 0;
+	bool skip_flip = false;
 
 	for (i = 0; i < DC_N_WINDOWS; i++) {
 		struct tegra_dc_ext_flip_win *flip_win = &data->win[i];
 		int index = flip_win->attr.index;
 		struct tegra_dc_win *win;
 		struct tegra_dc_ext_win *ext_win;
-		bool old_ena, new_ena;
 
 		if (index < 0)
 			continue;
@@ -266,28 +259,36 @@ static void tegra_dc_ext_flip_worker(struct work_struct *work)
 		win = tegra_dc_get_window(ext->dc, index);
 		ext_win = &ext->win[index];
 
-		win->swap_countdown = flip_win->attr.swap_interval;
+		if (!(atomic_dec_and_test(&ext_win->nr_pending_flips)) &&
+			(flip_win->attr.flags & TEGRA_DC_EXT_FLIP_FLAG_CURSOR))
+			skip_flip = true;
 
-		old_ena = ext->win[index].enabled;
-		new_ena = flip_win->handle != NULL;
-		if (old_ena != new_ena) {
-			if (new_ena)
-				process_window_change(ext, 1);
-			else
-				nr_disable++;
+		if (win->flags & TEGRA_WIN_FLAG_ENABLED) {
+			int j;
+			for (j = 0; j < TEGRA_DC_NUM_PLANES; j++) {
+				if (skip_flip)
+					old_handle = flip_win->handle[j];
+				else
+					old_handle = ext_win->cur_handle[j];
+
+				if (!old_handle)
+					continue;
+
+				unpin_handles[nr_unpin++] = old_handle;
+			}
 		}
-		ext->win[index].enabled = new_ena;
 
-		if (old_ena && ext_win->cur_handle)
-			unpin_handles[nr_unpin++] = ext_win->cur_handle;
-
-		tegra_dc_ext_set_windowattr(ext, win, &data->win[i]);
+		if (!skip_flip)
+			tegra_dc_ext_set_windowattr(ext, win, &data->win[i]);
 
 		wins[nr_win++] = win;
 	}
 
-	tegra_dc_update_windows(wins, nr_win);
-	tegra_dc_sync_windows(wins, nr_win);
+	if (!skip_flip) {
+		tegra_dc_update_windows(wins, nr_win);
+		/* TODO: implement swapinterval here */
+		tegra_dc_sync_windows(wins, nr_win);
+	}
 
 	for (i = 0; i < DC_N_WINDOWS; i++) {
 		struct tegra_dc_ext_flip_win *flip_win = &data->win[i];
@@ -306,9 +307,6 @@ static void tegra_dc_ext_flip_worker(struct work_struct *work)
 		nvmap_free(ext->nvmap, unpin_handles[i]);
 	}
 
-	if (nr_disable)
-		process_window_change(ext, -nr_disable);
-
 	kfree(data);
 }
 
@@ -316,16 +314,25 @@ static int lock_windows_for_flip(struct tegra_dc_ext_user *user,
 				 struct tegra_dc_ext_flip *args)
 {
 	struct tegra_dc_ext *ext = user->ext;
+	u8 idx_mask = 0;
 	int i;
 
 	for (i = 0; i < DC_N_WINDOWS; i++) {
 		int index = args->win[i].index;
-		struct tegra_dc_ext_win *win;
 
 		if (index < 0)
 			continue;
 
-		win = &ext->win[index];
+		idx_mask |= BIT(index);
+	}
+
+	for (i = 0; i < DC_N_WINDOWS; i++) {
+		struct tegra_dc_ext_win *win;
+
+		if (!(idx_mask & BIT(i)))
+			continue;
+
+		win = &ext->win[i];
 
 		mutex_lock(&win->lock);
 
@@ -337,12 +344,10 @@ static int lock_windows_for_flip(struct tegra_dc_ext_user *user,
 
 fail_unlock:
 	do {
-		int index = args->win[i].index;
-
-		if (index < 0)
+		if (!(idx_mask & BIT(i)))
 			continue;
 
-		mutex_unlock(&ext->win[index].lock);
+		mutex_unlock(&ext->win[i].lock);
 	} while (i--);
 
 	return -EACCES;
@@ -352,6 +357,7 @@ static void unlock_windows_for_flip(struct tegra_dc_ext_user *user,
 				    struct tegra_dc_ext_flip *args)
 {
 	struct tegra_dc_ext *ext = user->ext;
+	u8 idx_mask = 0;
 	int i;
 
 	for (i = 0; i < DC_N_WINDOWS; i++) {
@@ -360,7 +366,14 @@ static void unlock_windows_for_flip(struct tegra_dc_ext_user *user,
 		if (index < 0)
 			continue;
 
-		mutex_unlock(&ext->win[index].lock);
+		idx_mask |= BIT(index);
+	}
+
+	for (i = DC_N_WINDOWS - 1; i >= 0; i--) {
+		if (!(idx_mask & BIT(i)))
+			continue;
+
+		mutex_unlock(&ext->win[i].lock);
 	}
 }
 
@@ -398,6 +411,11 @@ static int tegra_dc_ext_flip(struct tegra_dc_ext_user *user,
 	int work_index;
 	int i, ret = 0;
 
+#ifdef CONFIG_ANDROID
+	int index_check[DC_N_WINDOWS] = {0, };
+	int zero_index_id = 0;
+#endif
+
 	if (!user->nvmap)
 		return -EFAULT;
 
@@ -412,6 +430,21 @@ static int tegra_dc_ext_flip(struct tegra_dc_ext_user *user,
 	INIT_WORK(&data->work, tegra_dc_ext_flip_worker);
 	data->ext = ext;
 
+#ifdef CONFIG_ANDROID
+	for (i = 0; i < DC_N_WINDOWS; i++) {
+		index_check[i] = args->win[i].index;
+		if (index_check[i] == 0)
+			zero_index_id = i;
+	}
+
+	if (index_check[DC_N_WINDOWS - 1] != 0) {
+		struct tegra_dc_ext_flip_windowattr win_temp;
+		win_temp = args->win[DC_N_WINDOWS - 1];
+		args->win[DC_N_WINDOWS - 1] = args->win[zero_index_id];
+		args->win[zero_index_id] = win_temp;
+	}
+#endif
+
 	for (i = 0; i < DC_N_WINDOWS; i++) {
 		struct tegra_dc_ext_flip_win *flip_win = &data->win[i];
 		int index = args->win[i].index;
@@ -422,10 +455,34 @@ static int tegra_dc_ext_flip(struct tegra_dc_ext_user *user,
 			continue;
 
 		ret = tegra_dc_ext_pin_window(user, flip_win->attr.buff_id,
-					      &flip_win->handle,
+					      &flip_win->handle[TEGRA_DC_Y],
 					      &flip_win->phys_addr);
 		if (ret)
 			goto fail_pin;
+
+		if (flip_win->attr.buff_id_u) {
+			ret = tegra_dc_ext_pin_window(user,
+					      flip_win->attr.buff_id_u,
+					      &flip_win->handle[TEGRA_DC_U],
+					      &flip_win->phys_addr_u);
+			if (ret)
+				goto fail_pin;
+		} else {
+			flip_win->handle[TEGRA_DC_U] = NULL;
+			flip_win->phys_addr_u = 0;
+		}
+
+		if (flip_win->attr.buff_id_v) {
+			ret = tegra_dc_ext_pin_window(user,
+					      flip_win->attr.buff_id_v,
+					      &flip_win->handle[TEGRA_DC_V],
+					      &flip_win->phys_addr_v);
+			if (ret)
+				goto fail_pin;
+		} else {
+			flip_win->handle[TEGRA_DC_V] = NULL;
+			flip_win->phys_addr_v = 0;
+		}
 	}
 
 	ret = lock_windows_for_flip(user, args);
@@ -440,10 +497,15 @@ static int tegra_dc_ext_flip(struct tegra_dc_ext_user *user,
 	for (i = 0; i < DC_N_WINDOWS; i++) {
 		u32 syncpt_max;
 		int index = args->win[i].index;
+		struct tegra_dc_win *win;
+		struct tegra_dc_ext_win *ext_win;
 
 		if (index < 0)
 			continue;
 
+		win = tegra_dc_get_window(ext->dc, index);
+		ext_win = &ext->win[index];
+
 		syncpt_max = tegra_dc_incr_syncpt_max(ext->dc, index);
 
 		data->win[i].syncpt_max = syncpt_max;
@@ -455,6 +517,8 @@ static int tegra_dc_ext_flip(struct tegra_dc_ext_user *user,
 		args->post_syncpt_val = syncpt_max;
 		args->post_syncpt_id = tegra_dc_get_syncpt_id(ext->dc, index);
 		work_index = index;
+
+		atomic_inc(&ext->win[work_index].nr_pending_flips);
 	}
 	queue_work(ext->win[work_index].flip_wq, &data->work);
 
@@ -466,12 +530,15 @@ unlock:
 	unlock_windows_for_flip(user, args);
 
 fail_pin:
-	while (i--) {
-		if (!data->win[i].handle)
-			continue;
+	for (i = 0; i < DC_N_WINDOWS; i++) {
+		int j;
+		for (j = 0; j < TEGRA_DC_NUM_PLANES; j++) {
+			if (!data->win[i].handle[j])
+				continue;
 
-		nvmap_unpin(ext->nvmap, data->win[i].handle);
-		nvmap_free(ext->nvmap, data->win[i].handle);
+			nvmap_unpin(ext->nvmap, data->win[i].handle[j]);
+			nvmap_free(ext->nvmap, data->win[i].handle[j]);
+		}
 	}
 	kfree(data);
 
@@ -515,6 +582,73 @@ static int tegra_dc_ext_set_csc(struct tegra_dc_ext_user *user,
 	return 0;
 }
 
+static int set_lut_channel(u16 *channel_from_user,
+			   u8 *channel_to,
+			   u32 start,
+			   u32 len)
+{
+	int i;
+	u16 lut16bpp[256];
+
+	if (channel_from_user) {
+		if (copy_from_user(lut16bpp, channel_from_user, len<<1))
+			return 1;
+
+		for (i = 0; i < len; i++)
+			channel_to[start+i] = lut16bpp[i]>>8;
+	} else {
+		for (i = 0; i < len; i++)
+			channel_to[start+i] = start+i;
+	}
+
+	return 0;
+}
+
+static int tegra_dc_ext_set_lut(struct tegra_dc_ext_user *user,
+				struct tegra_dc_ext_lut *new_lut)
+{
+	int err;
+	unsigned int index = new_lut->win_index;
+	u32 start = new_lut->start;
+	u32 len = new_lut->len;
+
+	struct tegra_dc *dc = user->ext->dc;
+	struct tegra_dc_ext_win *ext_win;
+	struct tegra_dc_lut *lut;
+
+	if (index >= DC_N_WINDOWS)
+		return -EINVAL;
+
+	if ((start >= 256) || (len > 256) || ((start + len) > 256))
+		return -EINVAL;
+
+	ext_win = &user->ext->win[index];
+	lut = &dc->windows[index].lut;
+
+	mutex_lock(&ext_win->lock);
+
+	if (ext_win->user != user) {
+		mutex_unlock(&ext_win->lock);
+		return -EACCES;
+	}
+
+	err = set_lut_channel(new_lut->r, lut->r, start, len) |
+	      set_lut_channel(new_lut->g, lut->g, start, len) |
+	      set_lut_channel(new_lut->b, lut->b, start, len);
+
+	if (err) {
+		mutex_unlock(&ext_win->lock);
+		return -EFAULT;
+	}
+
+	tegra_dc_update_lut(dc, index,
+			new_lut->flags & TEGRA_DC_EXT_LUT_FLAGS_FBOVERRIDE);
+
+	mutex_unlock(&ext_win->lock);
+
+	return 0;
+}
+
 static u32 tegra_dc_ext_get_vblank_syncpt(struct tegra_dc_ext_user *user)
 {
 	struct tegra_dc *dc = user->ext->dc;
@@ -622,6 +756,16 @@ static long tegra_dc_ioctl(struct file *filp, unsigned int cmd,
 		return ret;
 	}
 
+	case TEGRA_DC_EXT_SET_LUT:
+	{
+		struct tegra_dc_ext_lut args;
+
+		if (copy_from_user(&args, user_arg, sizeof(args)))
+			return -EFAULT;
+
+		return tegra_dc_ext_set_lut(user, &args);
+	}
+
 	default:
 		return -EINVAL;
 	}
@@ -752,7 +896,6 @@ struct tegra_dc_ext *tegra_dc_ext_register(struct nvhost_device *ndev,
 		goto cleanup_nvmap;
 
 	mutex_init(&ext->cursor.lock);
-	mutex_init(&ext->enable_change_lock);
 
 	head_count++;
 
