@@ -6,6 +6,8 @@
  *         Colin Cross <ccross@android.com>
  *         Travis Geiselbrecht <travis@palm.com>
  *
+ * Copyright (C) 2010-2011 NVIDIA Corporation
+ *
  * This software is licensed under the terms of the GNU General Public
  * License version 2, as published by the Free Software Foundation, and
  * may be copied, distributed, and modified under those terms.
@@ -34,17 +36,21 @@
 
 #include <mach/dc.h>
 #include <mach/fb.h>
-#include <mach/nvhost.h>
+#include <linux/nvhost.h>
 #include <mach/nvmap.h>
 
 #include "host/dev.h"
 #include "nvmap/nvmap.h"
+#include "dc/dc_priv.h"
+
+/* Pad pitch to 16-byte boundary. */
+#define TEGRA_LINEAR_PITCH_ALIGNMENT 16
 
 struct tegra_fb_info {
 	struct tegra_dc_win	*win;
 	struct nvhost_device	*ndev;
 	struct fb_info		*info;
-	bool			in_use;
+	bool			valid;
 
 	struct resource		*fb_mem;
 
@@ -59,17 +65,27 @@ static int tegra_fb_check_var(struct fb_var_screeninfo *var,
 			      struct fb_info *info)
 {
 	struct tegra_fb_info *tegra_fb = info->par;
+	struct tegra_dc *dc = tegra_fb->win->dc;
+	struct tegra_dc_out_ops *ops = dc->out_ops;
 	struct fb_videomode mode;
 
-	if ((var->yres * var->xres * var->bits_per_pixel / 8) >
+	if ((var->yres * var->xres * var->bits_per_pixel / 8 * 2) >
 	    info->screen_size)
 		return -EINVAL;
 
-	fb_var_to_videomode(&mode, var);
-	if (!tegra_dc_mode_filter(tegra_fb->win->dc, &mode))
-		return -EINVAL;
-	/* tegra_dc_mode_filter may have modified the mode */
-	fb_videomode_to_var(var, &mode);
+	/* Apply mode filter for HDMI only -LVDS supports only fix mode */
+	if (ops && ops->mode_filter) {
+
+		fb_var_to_videomode(&mode, var);
+		if (!ops->mode_filter(dc, &mode))
+			return -EINVAL;
+
+		/* Mode filter may have modified the mode */
+		fb_videomode_to_var(var, &mode);
+	}
+
+	/* Double yres_virtual to allow double buffering through pan_display */
+	var->yres_virtual = var->yres * 2;
 
 	return 0;
 }
@@ -91,6 +107,7 @@ static int tegra_fb_set_par(struct fb_info *info)
 			var->blue.length = 8;
 			var->transp.offset = 24;
 			var->transp.length = 8;
+			tegra_fb->win->fmt = TEGRA_WIN_FMT_R8G8B8A8;
 			break;
 		case 16:
 			var->red.offset = 11;
@@ -99,38 +116,52 @@ static int tegra_fb_set_par(struct fb_info *info)
 			var->green.length = 6;
 			var->blue.offset = 0;
 			var->blue.length = 5;
+			tegra_fb->win->fmt = TEGRA_WIN_FMT_B5G6R5;
 			break;
 
 		default:
 			return -EINVAL;
 		}
 		info->fix.line_length = var->xres * var->bits_per_pixel / 8;
+		/* Pad the stride to 16-byte boundary. */
+		info->fix.line_length = round_up(info->fix.line_length,
+						TEGRA_LINEAR_PITCH_ALIGNMENT);
+		tegra_fb->win->stride = info->fix.line_length;
+		tegra_fb->win->stride_uv = 0;
+		tegra_fb->win->phys_addr_u = 0;
+		tegra_fb->win->phys_addr_v = 0;
 	}
 
 	if (var->pixclock) {
-		struct tegra_dc_mode mode;
+		bool stereo;
+		struct fb_videomode m;
 
-		mode.pclk = PICOS2KHZ(var->pixclock) * 1000;
-		mode.h_ref_to_sync = 1;
-		mode.v_ref_to_sync = 1;
-		mode.h_sync_width = var->hsync_len;
-		mode.v_sync_width = var->vsync_len;
-		mode.h_back_porch = var->left_margin;
-		mode.v_back_porch = var->upper_margin;
-		mode.h_active = var->xres;
-		mode.v_active = var->yres;
-		mode.h_front_porch = var->right_margin;
-		mode.v_front_porch = var->lower_margin;
+		fb_var_to_videomode(&m, var);
 
-		mode.flags = 0;
+		info->mode = (struct fb_videomode *)
+			fb_find_nearest_mode(&m, &info->modelist);
+		if (!info->mode) {
+			dev_warn(&tegra_fb->ndev->dev, "can't match video mode\n");
+			return -EINVAL;
+		}
 
-		if (!(var->sync & FB_SYNC_HOR_HIGH_ACT))
-			mode.flags |= TEGRA_DC_MODE_FLAG_NEG_H_SYNC;
+		/*
+		 * only enable stereo if the mode supports it and
+		 * client requests it
+		 */
+		stereo = !!(var->vmode & info->mode->vmode &
+#ifndef CONFIG_TEGRA_HDMI_74MHZ_LIMIT
+					FB_VMODE_STEREO_FRAME_PACK);
+#else
+					FB_VMODE_STEREO_LEFT_RIGHT);
+#endif
 
-		if (!(var->sync & FB_SYNC_VERT_HIGH_ACT))
-			mode.flags |= TEGRA_DC_MODE_FLAG_NEG_V_SYNC;
+		tegra_dc_set_fb_mode(tegra_fb->win->dc, info->mode, stereo);
 
-		tegra_dc_set_mode(tegra_fb->win->dc, &mode);
+		tegra_fb->win->w.full = dfixed_const(info->mode->xres);
+		tegra_fb->win->h.full = dfixed_const(info->mode->yres);
+		tegra_fb->win->out_w = info->mode->xres;
+		tegra_fb->win->out_h = info->mode->yres;
 	}
 	return 0;
 }
@@ -161,41 +192,70 @@ static int tegra_fb_setcolreg(unsigned regno, unsigned red, unsigned green,
 	return 0;
 }
 
-static int tegra_fb_blank(int blank, struct fb_info *info)
+
+static int tegra_fb_setcmap(struct fb_cmap *cmap, struct fb_info *info)
 {
 	struct tegra_fb_info *tegra_fb = info->par;
+	struct tegra_dc *dc = tegra_fb->win->dc;
+	int i;
+	u16 *red = cmap->red;
+	u16 *green = cmap->green;
+	u16 *blue = cmap->blue;
+	int start = cmap->start;
 
-	switch (blank) {
-	case FB_BLANK_UNBLANK:
-		dev_dbg(&tegra_fb->ndev->dev, "unblank\n");
-		tegra_dc_enable(tegra_fb->win->dc);
-		return 0;
+	if (((unsigned)start > 255) || ((start + cmap->len) > 256))
+		return -EINVAL;
 
-	case FB_BLANK_NORMAL:
-	case FB_BLANK_VSYNC_SUSPEND:
-	case FB_BLANK_HSYNC_SUSPEND:
-	case FB_BLANK_POWERDOWN:
-		dev_dbg(&tegra_fb->ndev->dev, "blank\n");
-		tegra_dc_disable(tegra_fb->win->dc);
-		return 0;
+	if (info->fix.visual == FB_VISUAL_TRUECOLOR ||
+		info->fix.visual == FB_VISUAL_DIRECTCOLOR) {
+		/*
+		 * For now we are considering color schemes with
+		 * cmap->len <=16 as special case of basic color
+		 * scheme to support fbconsole.But for DirectColor
+		 * visuals(like the one we actually have, that include
+		 * a HW LUT),the way it's intended to work is that the
+		 * actual LUT HW is programmed to the intended values,
+		 * even for small color maps like those with 16 or fewer
+		 * entries. The pseudo_palette is then programmed to the
+		 * identity transform.
+		 */
+		if (cmap->len <= 16) {
+			/* Low-color schemes like fbconsole*/
+			u16 *transp = cmap->transp;
+			u_int vtransp = 0xffff;
 
-	default:
-		return -ENOTTY;
+			for (i = 0; i < cmap->len; i++) {
+				if (transp)
+					vtransp = *transp++;
+				if (tegra_fb_setcolreg(start++, *red++,
+					*green++, *blue++,
+					vtransp, info))
+						return -EINVAL;
+			}
+		} else {
+			/* High-color schemes*/
+			for (i = 0; i < cmap->len; i++) {
+				dc->fb_lut.r[start+i] = *red++ >> 8;
+				dc->fb_lut.g[start+i] = *green++ >> 8;
+				dc->fb_lut.b[start+i] = *blue++ >> 8;
+			}
+			tegra_dc_update_lut(dc, -1, -1);
+		}
 	}
+	return 0;
 }
 
+#if defined(CONFIG_FRAMEBUFFER_CONSOLE)
 static void tegra_fb_flip_win(struct tegra_fb_info *tegra_fb)
 {
 	struct tegra_dc_win *win = tegra_fb->win;
 	struct fb_info *info = tegra_fb->info;
 
-	if (!tegra_fb->in_use)
-		return;
-
 	win->x.full = dfixed_const(0);
 	win->y.full = dfixed_const(0);
 	win->w.full = dfixed_const(tegra_fb->xres);
 	win->h.full = dfixed_const(tegra_fb->yres);
+
 	/* TODO: set to output res dc */
 	win->out_x = 0;
 	win->out_y = 0;
@@ -206,8 +266,9 @@ static void tegra_fb_flip_win(struct tegra_fb_info *tegra_fb)
 		(info->var.yoffset * info->fix.line_length) +
 		(info->var.xoffset * (info->var.bits_per_pixel / 8));
 	win->virt_addr = info->screen_base;
-	win->offset_u = 0;
-	win->offset_v = 0;
+
+	win->phys_addr_u = 0;
+	win->phys_addr_v = 0;
 	win->stride = info->fix.line_length;
 	win->stride_uv = 0;
 
@@ -227,27 +288,61 @@ static void tegra_fb_flip_win(struct tegra_fb_info *tegra_fb)
 	tegra_dc_update_windows(&tegra_fb->win, 1);
 	tegra_dc_sync_windows(&tegra_fb->win, 1);
 }
+#endif
 
-static void tegra_fb_unflip_win(struct tegra_fb_info *tegra_fb)
+static int tegra_fb_blank(int blank, struct fb_info *info)
 {
-	struct tegra_dc_win *win = tegra_fb->win;
+	struct tegra_fb_info *tegra_fb = info->par;
 
-	win->flags &= ~TEGRA_WIN_FLAG_ENABLED;
+	switch (blank) {
+	case FB_BLANK_UNBLANK:
+		dev_dbg(&tegra_fb->ndev->dev, "unblank\n");
+		tegra_fb->win->flags = TEGRA_WIN_FLAG_ENABLED;
+		tegra_dc_enable(tegra_fb->win->dc);
+		return 0;
 
-	tegra_dc_update_windows(&tegra_fb->win, 1);
-	tegra_dc_sync_windows(&tegra_fb->win, 1);
+	case FB_BLANK_NORMAL:
+		dev_dbg(&tegra_fb->ndev->dev, "blank - normal\n");
+		tegra_dc_blank(tegra_fb->win->dc);
+		return 0;
+
+	case FB_BLANK_VSYNC_SUSPEND:
+	case FB_BLANK_HSYNC_SUSPEND:
+	case FB_BLANK_POWERDOWN:
+		dev_dbg(&tegra_fb->ndev->dev, "blank - powerdown\n");
+		tegra_dc_disable(tegra_fb->win->dc);
+		return 0;
+
+	default:
+		return -ENOTTY;
+	}
 }
 
 static int tegra_fb_pan_display(struct fb_var_screeninfo *var,
 				struct fb_info *info)
 {
 	struct tegra_fb_info *tegra_fb = info->par;
+	char __iomem *flush_start;
+	char __iomem *flush_end;
+	u32 addr;
 
-	info->var.xoffset = var->xoffset;
-	info->var.yoffset = var->yoffset;
+	if (!tegra_fb->win->cur_handle) {
+		flush_start = info->screen_base + (var->yoffset * info->fix.line_length);
+		flush_end = flush_start + (var->yres * info->fix.line_length);
 
-	tegra_fb_flip_win(tegra_fb);
-	/* TODO: update virt_addr */
+		info->var.xoffset = var->xoffset;
+		info->var.yoffset = var->yoffset;
+
+		addr = info->fix.smem_start + (var->yoffset * info->fix.line_length) +
+			(var->xoffset * (var->bits_per_pixel/8));
+
+		tegra_fb->win->phys_addr = addr;
+		tegra_fb->win->flags = TEGRA_WIN_FLAG_ENABLED;
+		tegra_fb->win->virt_addr = info->screen_base;
+
+		tegra_dc_update_windows(&tegra_fb->win, 1);
+		tegra_dc_sync_windows(&tegra_fb->win, 1);
+	}
 
 	return 0;
 }
@@ -287,12 +382,28 @@ static int tegra_fb_ioctl(struct fb_info *info, unsigned int cmd, unsigned long
 
 			if (i >= modedb.modedb_len)
 				break;
+
+			/* fb_videomode_to_var doesn't fill out all the members
+			   of fb_var_screeninfo */
+			memset(&var, 0x0, sizeof(var));
+
 			fb_videomode_to_var(&var, &modelist->mode);
 
 			if (copy_to_user((void __user *)&modedb.modedb[i],
 					 &var, sizeof(var)))
 				return -EFAULT;
 			i++;
+
+			if (var.vmode & FB_VMODE_STEREO_MASK) {
+				if (i >= modedb.modedb_len)
+					break;
+				var.vmode &= ~FB_VMODE_STEREO_MASK;
+				if (copy_to_user(
+					(void __user *)&modedb.modedb[i],
+					 &var, sizeof(var)))
+					return -EFAULT;
+				i++;
+			}
 		}
 		modedb.modedb_len = i;
 
@@ -307,11 +418,46 @@ static int tegra_fb_ioctl(struct fb_info *info, unsigned int cmd, unsigned long
 	return 0;
 }
 
+int tegra_fb_get_mode(struct tegra_dc *dc) {
+	return dc->fb->info->mode->refresh;
+}
+
+int tegra_fb_set_mode(struct tegra_dc *dc, int fps) {
+	size_t stereo;
+	struct list_head *pos;
+	struct fb_videomode *best_mode = NULL;
+	int curr_diff = INT_MAX; /* difference of best_mode refresh rate */
+	struct fb_modelist *modelist;
+	struct fb_info *info = dc->fb->info;
+
+	list_for_each(pos, &info->modelist) {
+		struct fb_videomode *mode;
+
+		modelist = list_entry(pos, struct fb_modelist, list);
+		mode = &modelist->mode;
+		if (fps <= mode->refresh && curr_diff > (mode->refresh - fps)) {
+			curr_diff = mode->refresh - fps;
+			best_mode = mode;
+		}
+	}
+	if (best_mode) {
+		info->mode = best_mode;
+		stereo = !!(info->var.vmode & info->mode->vmode &
+#ifndef CONFIG_TEGRA_HDMI_74MHZ_LIMIT
+				FB_VMODE_STEREO_FRAME_PACK);
+#else
+				FB_VMODE_STEREO_LEFT_RIGHT);
+#endif
+		return tegra_dc_set_fb_mode(dc, best_mode, stereo);
+	}
+	return -EIO;
+}
+
 static struct fb_ops tegra_fb_ops = {
 	.owner = THIS_MODULE,
 	.fb_check_var = tegra_fb_check_var,
 	.fb_set_par = tegra_fb_set_par,
-	.fb_setcolreg = tegra_fb_setcolreg,
+	.fb_setcmap = tegra_fb_setcmap,
 	.fb_blank = tegra_fb_blank,
 	.fb_pan_display = tegra_fb_pan_display,
 	.fb_fillrect = tegra_fb_fillrect,
@@ -326,7 +472,6 @@ void tegra_fb_update_monspecs(struct tegra_fb_info *fb_info,
 						  struct fb_videomode *mode))
 {
 	struct fb_event event;
-	struct fb_modelist *m;
 	int i;
 
 	mutex_lock(&fb_info->info->lock);
@@ -339,6 +484,12 @@ void tegra_fb_update_monspecs(struct tegra_fb_info *fb_info,
 		memset(&fb_info->info->monspecs, 0x0,
 		       sizeof(fb_info->info->monspecs));
 		memset(&mode, 0x0, sizeof(mode));
+
+		/*
+		 * reset video mode properties to prevent garbage being displayed on 'mode' device.
+		 */
+		fb_info->info->mode = (struct fb_videomode*) NULL;
+
 		tegra_dc_set_mode(fb_info->win->dc, &mode);
 		mutex_unlock(&fb_info->info->lock);
 		return;
@@ -346,28 +497,17 @@ void tegra_fb_update_monspecs(struct tegra_fb_info *fb_info,
 
 	memcpy(&fb_info->info->monspecs, specs,
 	       sizeof(fb_info->info->monspecs));
+	fb_info->info->mode = specs->modedb;
 
 	for (i = 0; i < specs->modedb_len; i++) {
-		if (!mode_filter ||
-		    mode_filter(fb_info->win->dc, &specs->modedb[i]))
+		if (mode_filter) {
+			if (mode_filter(fb_info->win->dc, &specs->modedb[i]))
+				fb_add_videomode(&specs->modedb[i],
+						 &fb_info->info->modelist);
+		} else {
 			fb_add_videomode(&specs->modedb[i],
 					 &fb_info->info->modelist);
-	}
-
-	if (list_empty(&fb_info->info->modelist)) {
-		struct tegra_dc_mode mode;
-		memset(&fb_info->info->var, 0x0, sizeof(fb_info->info->var));
-		memset(&mode, 0x0, sizeof(mode));
-		tegra_dc_set_mode(fb_info->win->dc, &mode);
-	} else {
-		/* in case the first mode was not matched */
-		m = list_first_entry(&fb_info->info->modelist, struct fb_modelist, list);
-		m->mode.flag |= FB_MODE_IS_FIRST;
-		fb_info->info->mode = (struct fb_videomode *)
-			fb_find_best_display(specs, &fb_info->info->modelist);
-
-		fb_videomode_to_var(&fb_info->info->var, fb_info->info->mode);
-		tegra_fb_set_par(fb_info->info);
+		}
 	}
 
 	event.info = fb_info->info;
@@ -375,19 +515,6 @@ void tegra_fb_update_monspecs(struct tegra_fb_info *fb_info,
 	mutex_unlock(&fb_info->info->lock);
 }
 
-void tegra_fb_transition(struct tegra_fb_info *tegra_fb, bool enable)
-{
-	if (!tegra_fb->fb_mem)
-		return;
-
-	tegra_fb->in_use = enable;
-	if (enable)
-		tegra_fb_flip_win(tegra_fb);
-	else
-		tegra_fb_unflip_win(tegra_fb);
-}
-EXPORT_SYMBOL(tegra_fb_transition);
-
 struct tegra_fb_info *tegra_fb_register(struct nvhost_device *ndev,
 					struct tegra_dc *dc,
 					struct tegra_fb_data *fb_data,
@@ -400,7 +527,6 @@ struct tegra_fb_info *tegra_fb_register(struct nvhost_device *ndev,
 	unsigned long fb_size = 0;
 	unsigned long fb_phys = 0;
 	int ret = 0;
-	const struct tegra_dc_mode *dc_mode;
 
 	win = tegra_dc_get_window(dc, fb_data->win);
 	if (!win) {
@@ -431,6 +557,7 @@ struct tegra_fb_info *tegra_fb_register(struct nvhost_device *ndev,
 			ret = -EBUSY;
 			goto err_free;
 		}
+		tegra_fb->valid = true;
 	}
 
 	info->fbops = &tegra_fb_ops;
@@ -446,32 +573,48 @@ struct tegra_fb_info *tegra_fb_register(struct nvhost_device *ndev,
 	info->fix.accel		= FB_ACCEL_NONE;
 	info->fix.smem_start	= fb_phys;
 	info->fix.smem_len	= fb_size;
+	info->fix.line_length = fb_data->xres * fb_data->bits_per_pixel / 8;
+	/* Pad the stride to 16-byte boundary. */
+	info->fix.line_length = round_up(info->fix.line_length,
+					TEGRA_LINEAR_PITCH_ALIGNMENT);
 
+	info->var.xres			= fb_data->xres;
+	info->var.yres			= fb_data->yres;
+	info->var.xres_virtual		= fb_data->xres;
+	info->var.yres_virtual		= fb_data->yres * 2;
 	info->var.bits_per_pixel	= fb_data->bits_per_pixel;
 	info->var.activate		= FB_ACTIVATE_VBL;
 	info->var.height		= tegra_dc_get_out_height(dc);
 	info->var.width			= tegra_dc_get_out_width(dc);
+	info->var.pixclock		= 0;
+	info->var.left_margin		= 0;
+	info->var.right_margin		= 0;
+	info->var.upper_margin		= 0;
+	info->var.lower_margin		= 0;
+	info->var.hsync_len		= 0;
+	info->var.vsync_len		= 0;
+	info->var.vmode			= FB_VMODE_NONINTERLACED;
 
-	dc_mode = tegra_dc_get_current_mode(dc);
-	if (dc_mode->pclk) {
-		struct fb_videomode mode;
+	win->x.full = dfixed_const(0);
+	win->y.full = dfixed_const(0);
+	win->w.full = dfixed_const(fb_data->xres);
+	win->h.full = dfixed_const(fb_data->yres);
+	/* TODO: set to output res dc */
+	win->out_x = 0;
+	win->out_y = 0;
+	win->out_w = fb_data->xres;
+	win->out_h = fb_data->yres;
+	win->z = 0;
+	win->phys_addr = fb_phys;
+	win->virt_addr = fb_base;
+	win->phys_addr_u = 0;
+	win->phys_addr_v = 0;
+	win->stride = info->fix.line_length;
+	win->stride_uv = 0;
+	win->flags = TEGRA_WIN_FLAG_ENABLED;
 
-		mode.pixclock		= KHZ2PICOS(dc_mode->pclk / 1000);
-		mode.xres		= dc_mode->h_active;
-		mode.yres		= dc_mode->v_active;
-		mode.left_margin	= dc_mode->h_back_porch;
-		mode.right_margin	= dc_mode->h_front_porch;
-		mode.upper_margin	= dc_mode->v_back_porch;
-		mode.lower_margin	= dc_mode->v_front_porch;
-		mode.hsync_len		= dc_mode->h_sync_width;
-		mode.vsync_len		= dc_mode->v_sync_width;
-		mode.vmode		= FB_VMODE_NONINTERLACED;
-
-		fb_videomode_to_var(&info->var, &mode);
-	}
-
-
-	tegra_fb->info = info;
+	if (fb_mem)
+		tegra_fb_set_par(info);
 
 	if (register_framebuffer(info)) {
 		dev_err(&ndev->dev, "failed to register framebuffer\n");
@@ -479,15 +622,29 @@ struct tegra_fb_info *tegra_fb_register(struct nvhost_device *ndev,
 		goto err_iounmap_fb;
 	}
 
-	if (fb_mem) {
-		tegra_fb->in_use = true;
-		tegra_fb_set_par(info);
-	}
+	tegra_fb->info = info;
 
 	dev_info(&ndev->dev, "probed\n");
 
-	if (fb_data->flags & TEGRA_FB_FLIP_ON_PROBE)
-		tegra_fb_flip_win(tegra_fb);
+	if (fb_data->flags & TEGRA_FB_FLIP_ON_PROBE) {
+		tegra_dc_update_windows(&tegra_fb->win, 1);
+		tegra_dc_sync_windows(&tegra_fb->win, 1);
+	}
+
+	if (dc->mode.pclk > 1000) {
+		struct tegra_dc_mode *mode = &dc->mode;
+
+		if (dc->out->flags & TEGRA_DC_OUT_ONE_SHOT_MODE)
+			info->var.pixclock = KHZ2PICOS(mode->rated_pclk / 1000);
+		else
+			info->var.pixclock = KHZ2PICOS(mode->pclk / 1000);
+		info->var.left_margin = mode->h_back_porch;
+		info->var.right_margin = mode->h_front_porch;
+		info->var.upper_margin = mode->v_back_porch;
+		info->var.lower_margin = mode->v_front_porch;
+		info->var.hsync_len = mode->h_sync_width;
+		info->var.vsync_len = mode->v_sync_width;
+	}
 
 	return tegra_fb;
 
