@@ -13,11 +13,17 @@
  * GNU General Public License for more details.
  *
  */
+#ifndef TEGRA_CAMERA_H
+#define TEGRA_CAMERA_H
+
+/* this is to enable VI pattern generator (Null Sensor) */
+#define TEGRA_CAMERA_ENABLE_PD2VI_CLK 0x1
 
 enum {
 	TEGRA_CAMERA_MODULE_ISP = 0,
 	TEGRA_CAMERA_MODULE_VI,
 	TEGRA_CAMERA_MODULE_CSI,
+	TEGRA_CAMERA_MODULE_MAX,
 };
 
 enum {
@@ -29,10 +35,22 @@ struct tegra_camera_clk_info {
 	uint id;
 	uint clk_id;
 	unsigned long rate;
+	uint flag;	/* to inform if any special bits need to enabled/disabled */
 };
 
+enum StereoCameraMode {
+	Main = 0x0,		/* Sets the default camera to Main */
+	StereoCameraMode_Left = 0x01,	/* the left camera is on. */
+	StereoCameraMode_Right = 0x02,	/* the right camera is on. */
+	StereoCameraMode_Stereo = 0x03,	/* both cameras are on. */
+	StereoCameraMode_Force32 = 0x7FFFFFFF
+};
+
+
 #define TEGRA_CAMERA_IOCTL_ENABLE		_IOWR('i', 1, uint)
 #define TEGRA_CAMERA_IOCTL_DISABLE		_IOWR('i', 2, uint)
 #define TEGRA_CAMERA_IOCTL_CLK_SET_RATE		\
 	_IOWR('i', 3, struct tegra_camera_clk_info)
 #define TEGRA_CAMERA_IOCTL_RESET		_IOWR('i', 4, uint)
+
+#endif
