@@ -30,11 +30,15 @@
 
 #include <asm/mach-types.h>
 
+#include <linux/clk.h>
 #include <linux/module.h>
 #include <linux/platform_device.h>
-#include <linux/regulator/consumer.h>
 #include <linux/slab.h>
 #include <linux/gpio.h>
+#include <linux/regulator/consumer.h>
+#ifdef CONFIG_SWITCH
+#include <linux/switch.h>
+#endif
 
 #include <mach/tegra_wm8903_pdata.h>
 
@@ -46,11 +50,13 @@
 
 #include "../codecs/wm8903.h"
 
-#include "tegra_das.h"
-#include "tegra_i2s.h"
 #include "tegra_pcm.h"
 #include "tegra_asoc_utils.h"
 
+#ifdef CONFIG_ARCH_TEGRA_2x_SOC
+#include "tegra20_das.h"
+#endif
+
 #define DRV_NAME "tegra-snd-wm8903"
 
 #define GPIO_SPKR_EN    BIT(0)
@@ -62,9 +68,13 @@
 struct tegra_wm8903 {
 	struct tegra_asoc_utils_data util_data;
 	struct tegra_wm8903_platform_data *pdata;
+	struct regulator *spk_reg;
+	struct regulator *dmic_reg;
 	int gpio_requested;
-	struct regulator *vdd_dmic; /* IS_ERR() -> no digital mic. */
-	bool vdd_dmic_enabled;
+#ifdef CONFIG_SWITCH
+	int jack_status;
+#endif
+	enum snd_soc_bias_level bias_level;
 };
 
 static int tegra_wm8903_hw_params(struct snd_pcm_substream *substream,
@@ -76,8 +86,10 @@ static int tegra_wm8903_hw_params(struct snd_pcm_substream *substream,
 	struct snd_soc_codec *codec = rtd->codec;
 	struct snd_soc_card *card = codec->card;
 	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
-	int srate, mclk;
+	int srate, mclk, i2s_daifmt;
 	int err;
+	struct clk *clk_m;
+	int rate;
 
 	srate = params_rate(params);
 	switch (srate) {
@@ -90,29 +102,60 @@ static int tegra_wm8903_hw_params(struct snd_pcm_substream *substream,
 		mclk = 256 * srate;
 		break;
 	}
+
+
+
+	clk_m = clk_get_sys(NULL, "clk_m");
+	if (IS_ERR(clk_m)) {
+		dev_err(card->dev, "Can't retrieve clk clk_m\n");
+		err = PTR_ERR(clk_m);
+		return err;
+	}
+	rate = clk_get_rate(clk_m);
+	printk("extern1 rate=%d\n",rate);
+
+#if TEGRA30_I2S_MASTER_PLAYBACK
 	/* FIXME: Codec only requires >= 3MHz if OSR==0 */
 	while (mclk < 6000000)
 		mclk *= 2;
 
+	i2s_daifmt = SND_SOC_DAIFMT_NB_NF |
+		     SND_SOC_DAIFMT_CBS_CFS;
+#else
+	mclk = rate;
+
+	i2s_daifmt = SND_SOC_DAIFMT_NB_NF |
+		     SND_SOC_DAIFMT_CBM_CFM;
+#endif
+
+
 	err = tegra_asoc_utils_set_rate(&machine->util_data, srate, mclk);
 	if (err < 0) {
-		dev_err(card->dev, "Can't configure clocks\n");
-		return err;
+		if (!(machine->util_data.set_mclk % mclk))
+			mclk = machine->util_data.set_mclk;
+		else {
+			dev_err(card->dev, "Can't configure clocks\n");
+			return err;
+		}
 	}
 
-	err = snd_soc_dai_set_fmt(codec_dai,
-					SND_SOC_DAIFMT_I2S |
-					SND_SOC_DAIFMT_NB_NF |
-					SND_SOC_DAIFMT_CBS_CFS);
+	tegra_asoc_utils_lock_clk_rate(&machine->util_data, 1);
+
+	/* Use DSP mode for mono on Tegra20 */
+	if ((params_channels(params) != 2) &&
+	    (machine_is_ventana() || machine_is_harmony() ||
+	    machine_is_kaen() || machine_is_aebl()))
+		i2s_daifmt |= SND_SOC_DAIFMT_DSP_A;
+	else
+		i2s_daifmt |= SND_SOC_DAIFMT_I2S;
+
+	err = snd_soc_dai_set_fmt(codec_dai, i2s_daifmt);
 	if (err < 0) {
 		dev_err(card->dev, "codec_dai fmt not set\n");
 		return err;
 	}
 
-	err = snd_soc_dai_set_fmt(cpu_dai,
-					SND_SOC_DAIFMT_I2S |
-					SND_SOC_DAIFMT_NB_NF |
-					SND_SOC_DAIFMT_CBS_CFS);
+	err = snd_soc_dai_set_fmt(cpu_dai, i2s_daifmt);
 	if (err < 0) {
 		dev_err(card->dev, "cpu_dai fmt not set\n");
 		return err;
@@ -125,21 +168,166 @@ static int tegra_wm8903_hw_params(struct snd_pcm_substream *substream,
 		return err;
 	}
 
+#ifdef CONFIG_ARCH_TEGRA_2x_SOC
+	err = tegra20_das_connect_dac_to_dap(TEGRA20_DAS_DAP_SEL_DAC1,
+					TEGRA20_DAS_DAP_ID_1);
+	if (err < 0) {
+		dev_err(card->dev, "failed to set dap-dac path\n");
+		return err;
+	}
+
+	err = tegra20_das_connect_dap_to_dac(TEGRA20_DAS_DAP_ID_1,
+					TEGRA20_DAS_DAP_SEL_DAC1);
+	if (err < 0) {
+		dev_err(card->dev, "failed to set dac-dap path\n");
+		return err;
+	}
+#endif
+	return 0;
+}
+
+static int tegra_bt_sco_hw_params(struct snd_pcm_substream *substream,
+					struct snd_pcm_hw_params *params)
+{
+	struct snd_soc_pcm_runtime *rtd = substream->private_data;
+	struct snd_soc_dai *cpu_dai = rtd->cpu_dai;
+	struct snd_soc_card *card = rtd->card;
+	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
+	int srate, mclk, min_mclk;
+	int err;
+
+	srate = params_rate(params);
+	switch (srate) {
+	case 11025:
+	case 22050:
+	case 44100:
+	case 88200:
+		mclk = 11289600;
+		break;
+	case 8000:
+	case 16000:
+	case 32000:
+	case 48000:
+	case 64000:
+	case 96000:
+		mclk = 12288000;
+		break;
+	default:
+		return -EINVAL;
+	}
+	min_mclk = 64 * srate;
+
+	err = tegra_asoc_utils_set_rate(&machine->util_data, srate, mclk);
+	if (err < 0) {
+		if (!(machine->util_data.set_mclk % min_mclk))
+			mclk = machine->util_data.set_mclk;
+		else {
+			dev_err(card->dev, "Can't configure clocks\n");
+			return err;
+		}
+	}
+
+	tegra_asoc_utils_lock_clk_rate(&machine->util_data, 1);
+
+	err = snd_soc_dai_set_fmt(cpu_dai,
+					SND_SOC_DAIFMT_DSP_A |
+					SND_SOC_DAIFMT_NB_NF |
+					SND_SOC_DAIFMT_CBS_CFS);
+	if (err < 0) {
+		dev_err(card->dev, "cpu_dai fmt not set\n");
+		return err;
+	}
+
+#ifdef CONFIG_ARCH_TEGRA_2x_SOC
+	err = tegra20_das_connect_dac_to_dap(TEGRA20_DAS_DAP_SEL_DAC2,
+					TEGRA20_DAS_DAP_ID_4);
+	if (err < 0) {
+		dev_err(card->dev, "failed to set dac-dap path\n");
+		return err;
+	}
+
+	err = tegra20_das_connect_dap_to_dac(TEGRA20_DAS_DAP_ID_4,
+					TEGRA20_DAS_DAP_SEL_DAC2);
+	if (err < 0) {
+		dev_err(card->dev, "failed to set dac-dap path\n");
+		return err;
+	}
+#endif
+	return 0;
+}
+
+static int tegra_spdif_hw_params(struct snd_pcm_substream *substream,
+					struct snd_pcm_hw_params *params)
+{
+	struct snd_soc_pcm_runtime *rtd = substream->private_data;
+	struct snd_soc_card *card = rtd->card;
+	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
+	int srate, mclk, min_mclk;
+	int err;
+
+	srate = params_rate(params);
+	switch (srate) {
+	case 11025:
+	case 22050:
+	case 44100:
+	case 88200:
+		mclk = 11289600;
+		break;
+	case 8000:
+	case 16000:
+	case 32000:
+	case 48000:
+	case 64000:
+	case 96000:
+		mclk = 12288000;
+		break;
+	default:
+		return -EINVAL;
+	}
+	min_mclk = 128 * srate;
+
+	err = tegra_asoc_utils_set_rate(&machine->util_data, srate, mclk);
+	if (err < 0) {
+		if (!(machine->util_data.set_mclk % min_mclk))
+			mclk = machine->util_data.set_mclk;
+		else {
+			dev_err(card->dev, "Can't configure clocks\n");
+			return err;
+		}
+	}
+
+	tegra_asoc_utils_lock_clk_rate(&machine->util_data, 1);
+
+	return 0;
+}
+
+static int tegra_hw_free(struct snd_pcm_substream *substream)
+{
+	struct snd_soc_pcm_runtime *rtd = substream->private_data;
+	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(rtd->card);
+
+	tegra_asoc_utils_lock_clk_rate(&machine->util_data, 0);
+
 	return 0;
 }
 
 static struct snd_soc_ops tegra_wm8903_ops = {
 	.hw_params = tegra_wm8903_hw_params,
+	.hw_free = tegra_hw_free,
+};
+
+static struct snd_soc_ops tegra_wm8903_bt_sco_ops = {
+	.hw_params = tegra_bt_sco_hw_params,
+	.hw_free = tegra_hw_free,
+};
+
+static struct snd_soc_ops tegra_spdif_ops = {
+	.hw_params = tegra_spdif_hw_params,
+	.hw_free = tegra_hw_free,
 };
 
 static struct snd_soc_jack tegra_wm8903_hp_jack;
-
-static struct snd_soc_jack_pin tegra_wm8903_hp_jack_pins[] = {
-	{
-		.pin = "Headphone Jack",
-		.mask = SND_JACK_HEADPHONE,
-	},
-};
+static struct snd_soc_jack tegra_wm8903_mic_jack;
 
 static struct snd_soc_jack_gpio tegra_wm8903_hp_jack_gpio = {
 	.name = "headphone detect",
@@ -148,7 +336,63 @@ static struct snd_soc_jack_gpio tegra_wm8903_hp_jack_gpio = {
 	.invert = 1,
 };
 
-static struct snd_soc_jack tegra_wm8903_mic_jack;
+#ifdef CONFIG_SWITCH
+/* These values are copied from Android WiredAccessoryObserver */
+enum headset_state {
+	BIT_NO_HEADSET = 0,
+	BIT_HEADSET = (1 << 0),
+	BIT_HEADSET_NO_MIC = (1 << 1),
+};
+
+static struct switch_dev tegra_wm8903_headset_switch = {
+	.name = "h2w",
+};
+
+static int tegra_wm8903_jack_notifier(struct notifier_block *self,
+			      unsigned long action, void *dev)
+{
+	struct snd_soc_jack *jack = dev;
+	struct snd_soc_codec *codec = jack->codec;
+	struct snd_soc_card *card = codec->card;
+	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
+	enum headset_state state = BIT_NO_HEADSET;
+
+	if (jack == &tegra_wm8903_hp_jack) {
+		machine->jack_status &= ~SND_JACK_HEADPHONE;
+		machine->jack_status |= (action & SND_JACK_HEADPHONE);
+	} else {
+		machine->jack_status &= ~SND_JACK_MICROPHONE;
+		machine->jack_status |= (action & SND_JACK_MICROPHONE);
+	}
+
+	switch (machine->jack_status) {
+	case SND_JACK_HEADPHONE:
+		state = BIT_HEADSET_NO_MIC;
+		break;
+	case SND_JACK_HEADSET:
+		state = BIT_HEADSET;
+		break;
+	case SND_JACK_MICROPHONE:
+		/* mic: would not report */
+	default:
+		state = BIT_NO_HEADSET;
+	}
+
+	switch_set_state(&tegra_wm8903_headset_switch, state);
+
+	return NOTIFY_OK;
+}
+
+static struct notifier_block tegra_wm8903_jack_detect_nb = {
+	.notifier_call = tegra_wm8903_jack_notifier,
+};
+#else
+static struct snd_soc_jack_pin tegra_wm8903_hp_jack_pins[] = {
+	{
+		.pin = "Headphone Jack",
+		.mask = SND_JACK_HEADPHONE,
+	},
+};
 
 static struct snd_soc_jack_pin tegra_wm8903_mic_jack_pins[] = {
 	{
@@ -156,6 +400,7 @@ static struct snd_soc_jack_pin tegra_wm8903_mic_jack_pins[] = {
 		.mask = SND_JACK_MICROPHONE,
 	},
 };
+#endif
 
 static int tegra_wm8903_event_int_spk(struct snd_soc_dapm_widget *w,
 					struct snd_kcontrol *k, int event)
@@ -165,6 +410,13 @@ static int tegra_wm8903_event_int_spk(struct snd_soc_dapm_widget *w,
 	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
 	struct tegra_wm8903_platform_data *pdata = machine->pdata;
 
+	if (machine->spk_reg) {
+		if (SND_SOC_DAPM_EVENT_ON(event))
+			regulator_enable(machine->spk_reg);
+		else
+			regulator_disable(machine->spk_reg);
+	}
+
 	if (!(machine->gpio_requested & GPIO_SPKR_EN))
 		return 0;
 
@@ -191,39 +443,60 @@ static int tegra_wm8903_event_hp(struct snd_soc_dapm_widget *w,
 	return 0;
 }
 
-static int wm8903_event_dmic(struct snd_soc_dapm_widget *w,
-			     struct snd_kcontrol *k, int event)
+static int tegra_wm8903_event_int_mic(struct snd_soc_dapm_widget *w,
+					struct snd_kcontrol *k, int event)
 {
 	struct snd_soc_dapm_context *dapm = w->dapm;
 	struct snd_soc_card *card = dapm->card;
 	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
-	bool new_enabled;
-	int ret;
+	struct tegra_wm8903_platform_data *pdata = machine->pdata;
 
-	if (IS_ERR(machine->vdd_dmic))
-	    return 0;
+	if (machine->dmic_reg) {
+		if (SND_SOC_DAPM_EVENT_ON(event))
+			regulator_enable(machine->dmic_reg);
+		else
+			regulator_disable(machine->dmic_reg);
+	}
 
-	new_enabled = !!SND_SOC_DAPM_EVENT_ON(event);
-	if (machine->vdd_dmic_enabled == new_enabled)
-	    return 0;
+	if (!(machine->gpio_requested & GPIO_INT_MIC_EN))
+		return 0;
 
-	if (new_enabled)
-	    ret = regulator_enable(machine->vdd_dmic);
-	else
-	    ret = regulator_disable(machine->vdd_dmic);
+	gpio_set_value_cansleep(pdata->gpio_int_mic_en,
+				SND_SOC_DAPM_EVENT_ON(event));
 
-	if (!ret)
-	    machine->vdd_dmic_enabled = new_enabled;
-
-	return ret;
+	return 0;
 }
 
+static int tegra_wm8903_event_ext_mic(struct snd_soc_dapm_widget *w,
+					struct snd_kcontrol *k, int event)
+{
+	struct snd_soc_dapm_context *dapm = w->dapm;
+	struct snd_soc_card *card = dapm->card;
+	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
+	struct tegra_wm8903_platform_data *pdata = machine->pdata;
 
-static const struct snd_soc_dapm_widget tegra_wm8903_dapm_widgets[] = {
+	if (!(machine->gpio_requested & GPIO_EXT_MIC_EN))
+		return 0;
+
+	gpio_set_value_cansleep(pdata->gpio_ext_mic_en,
+				SND_SOC_DAPM_EVENT_ON(event));
+
+	return 0;
+}
+
+static const struct snd_soc_dapm_widget cardhu_dapm_widgets[] = {
+	SND_SOC_DAPM_SPK("Int Spk", tegra_wm8903_event_int_spk),
+	SND_SOC_DAPM_HP("Headphone Jack", tegra_wm8903_event_hp),
+	SND_SOC_DAPM_LINE("LineOut", NULL),
+	SND_SOC_DAPM_MIC("Mic Jack", tegra_wm8903_event_ext_mic),
+	SND_SOC_DAPM_MIC("Int Mic", tegra_wm8903_event_int_mic),
+	SND_SOC_DAPM_LINE("Line In", NULL),
+};
+
+static const struct snd_soc_dapm_widget tegra_wm8903_default_dapm_widgets[] = {
 	SND_SOC_DAPM_SPK("Int Spk", tegra_wm8903_event_int_spk),
 	SND_SOC_DAPM_HP("Headphone Jack", tegra_wm8903_event_hp),
 	SND_SOC_DAPM_MIC("Mic Jack", NULL),
-	SND_SOC_DAPM_MIC("Digital Mic", wm8903_event_dmic),
 };
 
 static const struct snd_soc_dapm_route harmony_audio_map[] = {
@@ -237,6 +510,24 @@ static const struct snd_soc_dapm_route harmony_audio_map[] = {
 	{"IN1L", NULL, "Mic Bias"},
 };
 
+static const struct snd_soc_dapm_route cardhu_audio_map[] = {
+	{"Headphone Jack", NULL, "HPOUTR"},
+	{"Headphone Jack", NULL, "HPOUTL"},
+	{"Int Spk", NULL, "ROP"},
+	{"Int Spk", NULL, "RON"},
+	{"Int Spk", NULL, "LOP"},
+	{"Int Spk", NULL, "LON"},
+	{"LineOut", NULL, "LINEOUTL"},
+	{"LineOut", NULL, "LINEOUTR"},
+	{"Mic Bias", NULL, "Mic Jack"},
+	{"IN1L", NULL, "Mic Bias"},
+	{"Mic Bias", NULL, "Int Mic"},
+	{"IN1L", NULL, "Mic Bias"},
+	{"IN1R", NULL, "Mic Bias"},
+	{"IN3L", NULL, "Line In"},
+	{"IN3R", NULL, "Line In"},
+};
+
 static const struct snd_soc_dapm_route seaboard_audio_map[] = {
 	{"Headphone Jack", NULL, "HPOUTR"},
 	{"Headphone Jack", NULL, "HPOUTL"},
@@ -246,7 +537,6 @@ static const struct snd_soc_dapm_route seaboard_audio_map[] = {
 	{"Int Spk", NULL, "LON"},
 	{"Mic Bias", NULL, "Mic Jack"},
 	{"IN1R", NULL, "Mic Bias"},
-	{"DMICDAT", NULL, "Digital Mic"},
 };
 
 static const struct snd_soc_dapm_route kaen_audio_map[] = {
@@ -258,7 +548,6 @@ static const struct snd_soc_dapm_route kaen_audio_map[] = {
 	{"Int Spk", NULL, "LON"},
 	{"Mic Bias", NULL, "Mic Jack"},
 	{"IN2R", NULL, "Mic Bias"},
-	{"DMICDAT", NULL, "Digital Mic"},
 };
 
 static const struct snd_soc_dapm_route aebl_audio_map[] = {
@@ -268,142 +557,21 @@ static const struct snd_soc_dapm_route aebl_audio_map[] = {
 	{"Int Spk", NULL, "LINEOUTL"},
 	{"Mic Bias", NULL, "Mic Jack"},
 	{"IN1R", NULL, "Mic Bias"},
-	{"DMICDAT", NULL, "Digital Mic"},
 };
 
-static const struct snd_kcontrol_new tegra_wm8903_controls[] = {
+static const struct snd_kcontrol_new cardhu_controls[] = {
+	SOC_DAPM_PIN_SWITCH("Int Spk"),
+	SOC_DAPM_PIN_SWITCH("Headphone Jack"),
+	SOC_DAPM_PIN_SWITCH("LineOut"),
+	SOC_DAPM_PIN_SWITCH("Mic Jack"),
+	SOC_DAPM_PIN_SWITCH("Int Mic"),
+	SOC_DAPM_PIN_SWITCH("Line In"),
+};
+
+static const struct snd_kcontrol_new tegra_wm8903_default_controls[] = {
 	SOC_DAPM_PIN_SWITCH("Int Spk"),
 };
 
-/* tegra_wm8903_disconnect_pins
- *
- *   Disconnect pins which are not actually connected.
- *
- *   This implementation is easy to reason about, easy to maintain,
- *   and very easy to extend.
- *
- *   Unless specifically connected on a board, each pin will be marked
- *   as disconnected.
- *
- *   Add a New Machine
- *
- *      Add new machine by adding elements to the MACHINES macro.
- *
- *   Add a new Pin:
- *
- *      Insert the pin name into the 'wm8903_connected_pins' array.
- *      See 'Connect a Pin' to connect the pin.
- *
- *   Connect a Pin:
- *
- *      To connect a pin on a particular machine, use the CONNECT()
- *      macro.
- *
- *  Note: There is one distasteful artifact in this system, which I
- *        believe is acceptable due to the other advantages (listed
- *        above) of this implementation.  Notably, the expansion of
- *        'MACHINES' in the assignment to 'disconnect' below looks
- *        like syntactically incorrect code.
- */
-static void tegra_wm8903_disconnect_pins(struct snd_soc_card *card,
-					 struct snd_soc_dapm_context *dapm)
-{
-#define MACHINES                                \
-	M(aebl)                                 \
-	M(asymptote)                            \
-	M(harmony)                              \
-	M(kaen)                                 \
-	M(seaboard)                             \
-	M(ventana)
-
-#define M(_x) mn_##_x,
-	enum machine_names {
-		MACHINES
-		N_MACHINES
-	};
-#undef M
-	static const struct wm8903_connected_pins {
-		const char *name;
-		bool connected[N_MACHINES];
-	} pins[] = {
-#define CONNECT(_machine) .connected[mn_##_machine] = true
-		{
-			.name = "IN1L",
-			CONNECT(harmony),
-			CONNECT(ventana),
-		},
-		{
-			.name = "IN1R",
-			CONNECT(seaboard),
-			CONNECT(aebl),
-		},
-		{
-			.name = "IN2L",
-		},
-		{
-			.name = "IN2R",
-			CONNECT(kaen),
-		},
-		{
-			.name = "IN3L",
-		},
-		{
-			.name = "IN3R",
-		},
-		{
-			.name = "LON",
-			CONNECT(harmony),
-			CONNECT(ventana),
-			CONNECT(seaboard),
-			CONNECT(kaen),
-			CONNECT(asymptote),
-		},
-		{
-			.name = "RON",
-			CONNECT(harmony),
-			CONNECT(ventana),
-			CONNECT(seaboard),
-			CONNECT(kaen),
-			CONNECT(asymptote),
-		},
-		{
-			.name = "ROP",
-			CONNECT(harmony),
-			CONNECT(ventana),
-			CONNECT(seaboard),
-			CONNECT(kaen),
-			CONNECT(asymptote),
-		},
-		{
-			.name = "LOP",
-			CONNECT(harmony),
-			CONNECT(ventana),
-			CONNECT(seaboard),
-			CONNECT(kaen),
-			CONNECT(asymptote),
-		},
-		{
-			.name = "LINEOUTR",
-			CONNECT(aebl),
-		},
-		{
-			.name = "LINEOUTL",
-			CONNECT(aebl),
-		},
-#undef CONNECT
-	};
-	unsigned i;
-
-	/* FIXME: Calculate automatically based on DAPM routes? */
-	for (i = 0; i < ARRAY_SIZE(pins); ++i) {
-#define M(_x) || (machine_is_##_x() && !pins[i].connected[mn_##_x])
-		const bool disconnect = false MACHINES;
-		if (disconnect)
-			snd_soc_dapm_nc_pin(dapm, pins[i].name);
-#undef M
-	}
-}
-
 static int tegra_wm8903_init(struct snd_soc_pcm_runtime *rtd)
 {
 	struct snd_soc_codec *codec = rtd->codec;
@@ -413,6 +581,8 @@ static int tegra_wm8903_init(struct snd_soc_pcm_runtime *rtd)
 	struct tegra_wm8903_platform_data *pdata = machine->pdata;
 	int ret;
 
+	machine->bias_level = SND_SOC_BIAS_STANDBY;
+
 	if (gpio_is_valid(pdata->gpio_spkr_en)) {
 		ret = gpio_request(pdata->gpio_spkr_en, "spkr_en");
 		if (ret) {
@@ -432,7 +602,7 @@ static int tegra_wm8903_init(struct snd_soc_pcm_runtime *rtd)
 		}
 		machine->gpio_requested |= GPIO_HP_MUTE;
 
-		gpio_direction_output(pdata->gpio_hp_mute, 0);
+		gpio_direction_output(pdata->gpio_hp_mute, 1);
 	}
 
 	if (gpio_is_valid(pdata->gpio_int_mic_en)) {
@@ -461,12 +631,16 @@ static int tegra_wm8903_init(struct snd_soc_pcm_runtime *rtd)
 
 	if (gpio_is_valid(pdata->gpio_hp_det)) {
 		tegra_wm8903_hp_jack_gpio.gpio = pdata->gpio_hp_det;
-		tegra_wm8903_hp_jack_gpio.invert = pdata->gpio_hp_invert;
 		snd_soc_jack_new(codec, "Headphone Jack", SND_JACK_HEADPHONE,
 				&tegra_wm8903_hp_jack);
+#ifndef CONFIG_SWITCH
 		snd_soc_jack_add_pins(&tegra_wm8903_hp_jack,
 					ARRAY_SIZE(tegra_wm8903_hp_jack_pins),
 					tegra_wm8903_hp_jack_pins);
+#else
+		snd_soc_jack_notifier_register(&tegra_wm8903_hp_jack,
+					&tegra_wm8903_jack_detect_nb);
+#endif
 		snd_soc_jack_add_gpios(&tegra_wm8903_hp_jack,
 					1,
 					&tegra_wm8903_hp_jack_gpio);
@@ -475,40 +649,139 @@ static int tegra_wm8903_init(struct snd_soc_pcm_runtime *rtd)
 
 	snd_soc_jack_new(codec, "Mic Jack", SND_JACK_MICROPHONE,
 			 &tegra_wm8903_mic_jack);
+#ifndef CONFIG_SWITCH
 	snd_soc_jack_add_pins(&tegra_wm8903_mic_jack,
 			      ARRAY_SIZE(tegra_wm8903_mic_jack_pins),
 			      tegra_wm8903_mic_jack_pins);
+#else
+	snd_soc_jack_notifier_register(&tegra_wm8903_mic_jack,
+				&tegra_wm8903_jack_detect_nb);
+#endif
 	wm8903_mic_detect(codec, &tegra_wm8903_mic_jack, SND_JACK_MICROPHONE,
-				0);
+			  machine_is_cardhu() ? SND_JACK_MICROPHONE : 0);
 
 	snd_soc_dapm_force_enable_pin(dapm, "Mic Bias");
 
-	tegra_wm8903_disconnect_pins(card, dapm);
+	/* FIXME: Calculate automatically based on DAPM routes? */
+	if (!machine_is_harmony() && !machine_is_ventana() &&
+	    !machine_is_cardhu())
+		snd_soc_dapm_nc_pin(dapm, "IN1L");
+	if (!machine_is_seaboard() && !machine_is_aebl() &&
+	    !machine_is_cardhu())
+		snd_soc_dapm_nc_pin(dapm, "IN1R");
+	snd_soc_dapm_nc_pin(dapm, "IN2L");
+	if (!machine_is_kaen())
+		snd_soc_dapm_nc_pin(dapm, "IN2R");
+	snd_soc_dapm_nc_pin(dapm, "IN3L");
+	snd_soc_dapm_nc_pin(dapm, "IN3R");
+
+	if (machine_is_aebl()) {
+		snd_soc_dapm_nc_pin(dapm, "LON");
+		snd_soc_dapm_nc_pin(dapm, "RON");
+		snd_soc_dapm_nc_pin(dapm, "ROP");
+		snd_soc_dapm_nc_pin(dapm, "LOP");
+	} else {
+		snd_soc_dapm_nc_pin(dapm, "LINEOUTR");
+		snd_soc_dapm_nc_pin(dapm, "LINEOUTL");
+	}
+
 	snd_soc_dapm_sync(dapm);
 
 	return 0;
 }
 
-static struct snd_soc_dai_link tegra_wm8903_dai = {
-	.name = "WM8903",
-	.stream_name = "WM8903 PCM",
-	.codec_name = "wm8903.0-001a",
-	.platform_name = "tegra-pcm-audio",
-	.cpu_dai_name = "tegra-i2s.0",
-	.codec_dai_name = "wm8903-hifi",
-	.init = tegra_wm8903_init,
-	.ops = &tegra_wm8903_ops,
+#ifdef WM8903_SET_BIAS_LEVEL
+static int tegra30_soc_set_bias_level(struct snd_soc_card *card,
+					enum snd_soc_bias_level level)
+{
+	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
+
+	if (machine->bias_level == SND_SOC_BIAS_OFF &&
+		level != SND_SOC_BIAS_OFF)
+		tegra_asoc_utils_clk_enable(&machine->util_data);
+
+	return 0;
+}
+
+static int tegra30_soc_set_bias_level_post(struct snd_soc_card *card,
+					enum snd_soc_bias_level level)
+{
+	struct tegra_wm8903 *machine = snd_soc_card_get_drvdata(card);
+
+	if (machine->bias_level != SND_SOC_BIAS_OFF &&
+		level == SND_SOC_BIAS_OFF)
+		tegra_asoc_utils_clk_disable(&machine->util_data);
+
+	machine->bias_level = level;
+
+	return 0 ;
+}
+#endif
+
+static struct snd_soc_dai_link tegra_wm8903_dai[] = {
+	{
+		.name = "WM8903",
+		.stream_name = "WM8903 PCM",
+		.codec_name = "wm8903.0-001a",
+		.platform_name = "tegra-pcm-audio",
+		.cpu_dai_name = "tegra20-i2s.0",
+		.codec_dai_name = "wm8903-hifi",
+		.init = tegra_wm8903_init,
+		.ops = &tegra_wm8903_ops,
+	},
+	{
+		.name = "SPDIF",
+		.stream_name = "SPDIF PCM",
+		.codec_name = "spdif-dit.0",
+		.platform_name = "tegra-pcm-audio",
+		.cpu_dai_name = "tegra20-spdif",
+		.codec_dai_name = "dit-hifi",
+		.ops = &tegra_spdif_ops,
+	},
+	{
+		.name = "BT-SCO",
+		.stream_name = "BT SCO PCM",
+		.codec_name = "spdif-dit.1",
+		.platform_name = "tegra-pcm-audio",
+		.cpu_dai_name = "tegra20-i2s.1",
+		.codec_dai_name = "dit-hifi",
+		.ops = &tegra_wm8903_bt_sco_ops,
+	},
 };
 
+static int tegra_wm8903_suspend_post(struct snd_soc_card *card)
+{
+	struct snd_soc_jack_gpio *gpio = &tegra_wm8903_hp_jack_gpio;
+
+	if (gpio_is_valid(gpio->gpio))
+		disable_irq(gpio_to_irq(gpio->gpio));
+
+	return 0;
+}
+
+static int tegra_wm8903_resume_pre(struct snd_soc_card *card)
+{
+	int val;
+	struct snd_soc_jack_gpio *gpio = &tegra_wm8903_hp_jack_gpio;
+
+	if (gpio_is_valid(gpio->gpio)) {
+		val = gpio_get_value(gpio->gpio);
+		val = gpio->invert ? !val : val;
+		snd_soc_jack_report(gpio->jack, val, gpio->report);
+		enable_irq(gpio_to_irq(gpio->gpio));
+	}
+
+	return 0;
+}
+
 static struct snd_soc_card snd_soc_tegra_wm8903 = {
 	.name = "tegra-wm8903",
-	.dai_link = &tegra_wm8903_dai,
-	.num_links = 1,
-
-	.controls = tegra_wm8903_controls,
-	.num_controls = ARRAY_SIZE(tegra_wm8903_controls),
-	.dapm_widgets = tegra_wm8903_dapm_widgets,
-	.num_dapm_widgets = ARRAY_SIZE(tegra_wm8903_dapm_widgets),
+	.dai_link = tegra_wm8903_dai,
+	.num_links = ARRAY_SIZE(tegra_wm8903_dai),
+	.suspend_post = tegra_wm8903_suspend_post,
+	.resume_pre = tegra_wm8903_resume_pre,
+	//.set_bias_level = tegra30_soc_set_bias_level,
+	//.set_bias_level_post = tegra30_soc_set_bias_level_post,
 };
 
 static __devinit int tegra_wm8903_driver_probe(struct platform_device *pdev)
@@ -532,32 +805,74 @@ static __devinit int tegra_wm8903_driver_probe(struct platform_device *pdev)
 
 	machine->pdata = pdata;
 
-	ret = tegra_asoc_utils_init(&machine->util_data, &pdev->dev);
+	ret = tegra_asoc_utils_init(&machine->util_data, &pdev->dev, card);
 	if (ret)
 		goto err_free_machine;
 
-	machine->vdd_dmic = regulator_get(&pdev->dev, "vdd_dmic");
-	if (IS_ERR(machine->vdd_dmic)) {
-		dev_info(&pdev->dev,
-			 "Unable to locate 'vdd_dmic', error %ld\n",
-			 PTR_ERR(machine->vdd_dmic));
+	if (machine_is_cardhu() || machine_is_ventana()) {
+		machine->spk_reg = regulator_get(&pdev->dev, "vdd_spk_amp");
+		if (IS_ERR(machine->spk_reg)) {
+			dev_info(&pdev->dev, "No speaker regulator found\n");
+			machine->spk_reg = 0;
+		}
 	}
 
+	if (machine_is_ventana()) {
+		machine->dmic_reg = regulator_get(&pdev->dev, "vdd_dmic");
+		if (IS_ERR(machine->dmic_reg)) {
+			dev_info(&pdev->dev, "No digital mic"
+						" regulator found\n");
+			machine->dmic_reg = 0;
+		}
+	}
+
+	if (machine_is_cardhu()) {
+		tegra_wm8903_dai[0].codec_name = "wm8903.4-001a",
+		tegra_wm8903_dai[0].cpu_dai_name = "tegra30-i2s.1";
+
+		tegra_wm8903_dai[1].cpu_dai_name = "tegra30-spdif";
+
+		tegra_wm8903_dai[2].cpu_dai_name = "tegra30-i2s.3";
+	}
+
+#ifdef CONFIG_SWITCH
+	/* Addd h2w swith class support */
+	ret = switch_dev_register(&tegra_wm8903_headset_switch);
+	if (ret < 0)
+		goto err_fini_utils;
+#endif
+
 	card->dev = &pdev->dev;
 	platform_set_drvdata(pdev, card);
 	snd_soc_card_set_drvdata(card, machine);
 
-	if (machine_is_harmony() || machine_is_ventana()) {
+	if (machine_is_cardhu() || machine_is_ventana()) {
+		card->controls = cardhu_controls;
+		card->num_controls = ARRAY_SIZE(cardhu_controls);
+
+		card->dapm_widgets = cardhu_dapm_widgets;
+		card->num_dapm_widgets = ARRAY_SIZE(cardhu_dapm_widgets);
+	} else {
+		card->controls = tegra_wm8903_default_controls;
+		card->num_controls = ARRAY_SIZE(tegra_wm8903_default_controls);
+
+		card->dapm_widgets = tegra_wm8903_default_dapm_widgets;
+		card->num_dapm_widgets = ARRAY_SIZE(tegra_wm8903_default_dapm_widgets);
+	}
+
+	if (machine_is_harmony()) {
 		card->dapm_routes = harmony_audio_map;
 		card->num_dapm_routes = ARRAY_SIZE(harmony_audio_map);
-	} else if (machine_is_seaboard() || machine_is_asymptote()) {
+	} else if (machine_is_ventana() || machine_is_cardhu()) {
+		card->dapm_routes = cardhu_audio_map;
+		card->num_dapm_routes = ARRAY_SIZE(cardhu_audio_map);
+	} else if (machine_is_seaboard()) {
 		card->dapm_routes = seaboard_audio_map;
 		card->num_dapm_routes = ARRAY_SIZE(seaboard_audio_map);
 	} else if (machine_is_kaen()) {
 		card->dapm_routes = kaen_audio_map;
 		card->num_dapm_routes = ARRAY_SIZE(kaen_audio_map);
 	} else {
-		BUG_ON(!machine_is_aebl());
 		card->dapm_routes = aebl_audio_map;
 		card->num_dapm_routes = ARRAY_SIZE(aebl_audio_map);
 	}
@@ -566,18 +881,28 @@ static __devinit int tegra_wm8903_driver_probe(struct platform_device *pdev)
 	if (ret) {
 		dev_err(&pdev->dev, "snd_soc_register_card failed (%d)\n",
 			ret);
-		goto err_fini_utils;
+		goto err_unregister_switch;
+	}
+
+	if (!card->instantiated) {
+		ret = -ENODEV;
+		dev_err(&pdev->dev, "snd_soc_register_card failed (%d)\n",
+			ret);
+		goto err_unregister_card;
 	}
 
 	return 0;
 
+err_unregister_card:
+	snd_soc_unregister_card(card);
+err_unregister_switch:
+#ifdef CONFIG_SWITCH
+	switch_dev_unregister(&tegra_wm8903_headset_switch);
+#endif
 err_fini_utils:
-	if (!IS_ERR(machine->vdd_dmic))
-		regulator_put(machine->vdd_dmic);
 	tegra_asoc_utils_fini(&machine->util_data);
 err_free_machine:
 	kfree(machine);
-	dev_err(&pdev->dev, "Failed to probe wm8903 (%d)\n", ret);
 	return ret;
 }
 
@@ -601,13 +926,18 @@ static int __devexit tegra_wm8903_driver_remove(struct platform_device *pdev)
 		gpio_free(pdata->gpio_spkr_en);
 	machine->gpio_requested = 0;
 
-	snd_soc_unregister_card(card);
+	if (machine->spk_reg)
+		regulator_put(machine->spk_reg);
+	if (machine->dmic_reg)
+		regulator_put(machine->dmic_reg);
 
-	if (!IS_ERR(machine->vdd_dmic))
-		regulator_put(machine->vdd_dmic);
+	snd_soc_unregister_card(card);
 
 	tegra_asoc_utils_fini(&machine->util_data);
 
+#ifdef CONFIG_SWITCH
+	switch_dev_unregister(&tegra_wm8903_headset_switch);
+#endif
 	kfree(machine);
 
 	return 0;
